'''
Created on Aug 21, 2014

@author: ajn
'''
import pyparticleest.interfaces as interfaces
import pyparticleest.utils.ckalman as kalman
import numpy
from pyparticleest.filter import ParticleApproximation, TrajectoryStep

class DegenerateDoubleInt(interfaces.FFBSiRS, interfaces.ParticleFilteringNonMarkov):
    """ Degenerate model of a double integrator """
    def __init__(self, Q, R, p0=0.0, v0=0.0):
        self.Q = numpy.asarray(Q).reshape((1, 1))
        self.R = numpy.asarray(R).reshape((1, 1))
        self.p0 = p0
        self.v0 = v0
        return

    def create_initial_estimate(self, N):
        part = numpy.empty((N, 2))
        part[:, 0] = self.p0
        part[:, 1] = self.v0
        return part

    def eval_logp_x0(self, particles, t):

        N = len(particles)
        res = numpy.zeros(N)
        # Assumes Px0 is either full rang or zero
        res[particles[:, 0] != self.p0] = -numpy.Inf
        res[particles[:, 1] != self.v0] = -numpy.Inf

        return res

    def sample_process_noise_full(self, ptraj, ancestors, ut, tt):
        """ Return sampled process noise for the non-linear states """
        N = len(ancestors)
        noise = numpy.sqrt(self.Q) * numpy.random.normal(size=(1, N))
        return noise.T

    def update_full(self, particles, traj, uvec, yvec, tvec, ancestors, noise):
        """ Update estimate using 'data' as input """
        particles[:, 0] = particles[:, 0] + particles[:, 1]
        particles[:, 1] = particles[:, 1] + noise[:, 0]
        return particles

    def measure_full(self, particles, traj, uvec, yvec, tvec, ancestors):
        """ Return the log-pdf value of the measurement """
        diff = yvec[-1] - particles[:, 0]
        lpy = kalman.lognormpdf_scalar(diff, self.R)
        return lpy

    def eval_1st_stage_weights(self, particles, u, y, t):
        part = numpy.copy(particles)
        noise = numpy.zeros_like(part)
        partn = self.update(part, u, t, noise)
        return self.measure(partn, y, t + 1)

    def logp_xnext_max(self, particles, u, t):
        return kalman.lognormpdf_scalar(numpy.zeros((1,)), self.Q)

    def logp_xnext(self, particles, next_part, u, t):
        f = numpy.copy(particles)
        f[:, 0] = f[:, 0] + f[:, 1]

        diff = next_part - f
        lpx = kalman.lognormpdf_scalar(diff[:, 1], self.Q)
        lpx[diff[:, 0] != 0.0] = -numpy.Inf
        return lpx

    def sample_smooth(self, part, ptraj, anc, future_trajs, find, ut, yt, tt, cur_ind):
        """ Implements the sample_smooth function for MixedNLGaussian models """
        return part

    def propose_smooth(self, partp, up, tp, ut, yt, tt, future_trajs):
        """ Sample from a distrubtion q(x_t | x_{t-1}, x_{t+1}, y_t) """
        # Trivial choice of q, discard y_T and x_{t+1}
        if (partp != None):
            noise = self.sample_process_noise(partp, up, tp)
            prop_part = numpy.copy(partp)
            prop_part = self.update(prop_part, up, tp, noise)
        else:
            prop_part = self.create_initial_estimate(future_trajs.shape[1])
        return prop_part

    def logp_proposal(self, prop_part, partp, up, tp, ut, yt, tt, future_trajs):
        """ Eval log q(x_t | x_{t-1}, x_{t+1}, y_t) """
        if (partp != None):
            return self.logp_xnext(partp, prop_part, up, tp)
        else:
            return self.eval_logp_x0(prop_part, t=tt[0])

    def copy_ind(self, particles, new_ind=None):
        if (new_ind != None):
            return numpy.copy(particles[new_ind])
        else:
            return numpy.copy(particles)

    def set_params(self, params):
        pass

    def post_smoothing(self, st):
        return self.pre_mhips_pass(st)

    def pre_mhips_pass(self, st):
        return st.traj

class MarginDoubleInt(interfaces.FFBSiNonMarkov,
                      interfaces.ParticleFilteringNonMarkov):
    """ Degenerate model of a double integrator """
    def __init__(self, Q, R, p0, v0):
        self.Q = numpy.asarray(Q).reshape((1, 1))
        self.R = numpy.asarray(R).reshape((1, 1))
        self.p0 = p0
        self.v0 = v0
        return

    def create_initial_estimate(self, N):
        part = numpy.empty((N, 2))
        part[:, 0] = self.v0
        part[:, 1] = self.p0
        return part

    def eval_logp_x0(self, particles, t):

        N = len(particles)
        res = numpy.zeros(N)
        # Assumes Px0 is either full rang or zero
        res[particles[:, 0] != self.v0] = -numpy.Inf

        return res

    def sample_process_noise_full(self, ptraj, ancestors, ut, tt):
        """ Return sampled process noise for the non-linear states """
        N = len(ancestors)
        noise = numpy.sqrt(self.Q) * numpy.random.normal(size=(1, N))
        return noise.T

    def update_full(self, particles, traj, uvec, yvec, tvec, ancestors, noise):
        """ Update estimate using 'data' as input """
        particles[:, 1] += particles[:, 0]
        particles[:, 0] += noise[:, 0]
        return particles

    def measure_full(self, particles, traj, uvec, yvec, tvec, ancestors):
        """ Return the log-pdf value of the measurement """
        diff = yvec[-1] - particles[:, 1]
        lpy = kalman.lognormpdf_scalar(diff, self.R)
        return lpy

    def eval_1st_stage_weights(self, particles, u, y, t):
        part = numpy.copy(particles)
        noise = numpy.zeros_like(part)
        partn = self.update(part, u, t, noise)
        return self.measure(partn, y, t + 1)

    def logp_xnext_full(self, part, past_trajs, pind, future_trajs, find, ut, yt, tt, cur_ind):
        numpart = len(part)
        next_part = future_trajs[0].pa.part[find]
        Nn = len(next_part)
        if (numpart > 1 and Nn == 1):
            next_part = numpy.repeat(next_part, numpart, 0)

        lpx = numpy.empty(numpart)
        for j in xrange(numpart):
            Nt = next_part[j, 2]
            gamma = next_part[j, 3]
            Ptplus = part[j, 0] + part[j, 1]
            lpx[j] = -0.5 / self.R * (Nt * (Ptplus ** 2.0) - 2 * Ptplus * gamma) - 0.5 / self.Q * (next_part[j, 0] - part[j, 0]) ** 2
        return lpx

    def logp_xnext_singlestep(self, part, past_trajs, pind, future_parts, find, ut, yt, tt, cur_ind):
        return kalman.lognormpdf_scalar(future_parts[find, 0] - part[:, 0], self.Q)

    def sample_smooth(self, part, ptraj, anc, future_trajs, find, ut, yt, tt, cur_ind):
        """ Implements the sample_smooth function for MixedNLGaussian models """
        M = len(part)
        spart = numpy.empty((M, 4))
        spart[:, :2] = part[:, :2]

        if (future_trajs != None):
            for j in xrange(M):
                Ntp = future_trajs[0].pa.part[find[j], 2]
                gamma = future_trajs[0].pa.part[find[j], 3] + yt[cur_ind] - Ntp * part[j, 0]
                spart[j, 2] = Ntp + 1
                spart[j, 3] = gamma
        else:
            # Store gamma and Nt for t=T-1
            spart[:, 2] = 1
            spart[:, 3] = yt[cur_ind]

        return spart

    def propose_smooth(self, partp, up, tp, ut, yt, tt, future_trajs):
        """ Sample from a distrubtion q(x_t | x_{t-1}, x_{t+1}, y_t) """
        # Trivial choice of q, discard y_T and x_{t+1}
        if (partp != None):
            noise = self.sample_process_noise(partp, up, tp)
            prop_part = numpy.copy(partp[:, :2])
            self.update(prop_part, up, tp, noise)
        else:
            M = future_trajs.shape[1]
            prop_part = self.create_initial_estimate(M)
        return prop_part

    def logp_proposal(self, prop_part, partp, up, tp, ut, yt, tt, future_trajs):
        """ Eval log q(x_t | x_{t-1}, x_{t+1}, y_t) """
        if (partp != None):
            return kalman.lognormpdf_scalar(prop_part[:, 0] - partp[:, 0], self.Q)
        else:
            return self.eval_logp_x0(prop_part, t=tt[0])

    def copy_ind(self, particles, new_ind=None):
        if (new_ind != None):
            return numpy.copy(particles[new_ind])
        else:
            return numpy.copy(particles)

    def set_params(self, params):
        pass

    def post_smoothing(self, st):
        return self.pre_mhips_pass(st)

    def pre_mhips_pass(self, st):
        T = len(st.traj)
        M = len(st.traj[0].pa.part)

        traj = numpy.empty((T,), dtype=object)

        tmp = numpy.empty((M, 2))
        tmp[:, 0] = st.traj[0].pa.part[:, 0]
        tmp[:, 1] = self.p0
        traj[0] = TrajectoryStep(ParticleApproximation(tmp), st.traj[0].ancestors)

        for i in xrange(1, len(traj)):
            tmp[:, 0] = st.traj[i].pa.part[:, 0]
            tmp[:, 1] = traj[i - 1].pa.part[:, 0] + traj[i - 1].pa.part[:, 1]
            traj[i] = TrajectoryStep(ParticleApproximation(tmp), st.traj[i].ancestors)

        return traj
